@@ -182,7 +182,7 @@ handle_output_destroy(struct wl_listener *listener, void *data)
 		wlr_scene_node_destroy(&output->layer_tree[i]->node);
 	}
 	wlr_scene_node_destroy(&output->layer_popup_tree->node);
-	wlr_scene_node_destroy(&output->osd_tree->node);
+	wlr_scene_node_destroy(&output->cycle_osd_tree->node);
 	wlr_scene_node_destroy(&output->session_lock_tree->node);
 	if (output->workspace_osd) {
 		wlr_scene_node_destroy(&output->workspace_osd->node);
@@ -414,6 +414,41 @@ configure_new_output(struct server *server, struct output *output)
 	server->pending_output_layout_change--;
 }
 
+static uint64_t
+get_unused_output_id_bit(struct server *server)
+{
+	uint64_t used_id_bits = 0;
+	struct output *output;
+	wl_list_for_each(output, &server->outputs, link) {
+		used_id_bits |= output->id_bit;
+	}
+
+	if (used_id_bits == UINT64_MAX) {
+		return 0;
+	}
+
+	uint64_t id_bit = server->next_output_id_bit;
+	/*
+	 * __builtin_popcountll() should be supported by GCC & clang.
+	 * If it causes portability issues, just remove the assert.
+	 */
+	assert(__builtin_popcountll(id_bit) == 1);
+
+	while ((id_bit & used_id_bits) != 0) {
+		id_bit = (id_bit << 1) | (id_bit >> 63); /* rotate left */
+	}
+
+	/*
+	 * The current implementation of view_update_outputs() isn't
+	 * robust against ID bit re-use. Save the next bit here so we
+	 * can cycle through all 64 available bits, making re-use less
+	 * frequent (on a best-effort basis).
+	 */
+	server->next_output_id_bit = (id_bit << 1) | (id_bit >> 63);
+
+	return id_bit;
+}
+
 static void
 handle_new_output(struct wl_listener *listener, void *data)
 {
@@ -437,6 +472,12 @@ handle_new_output(struct wl_listener *listener, void *data)
 		}
 	}
 
+	uint64_t id_bit = get_unused_output_id_bit(server);
+	if (!id_bit) {
+		wlr_log(WLR_ERROR, "Cannot add more than 64 outputs");
+		return;
+	}
+
 	if (wlr_output_is_wl(wlr_output)) {
 		char title[64];
 		snprintf(title, sizeof(title), "%s - %s", "labwc", wlr_output->name);
@@ -487,6 +528,7 @@ handle_new_output(struct wl_listener *listener, void *data)
 	output->wlr_output = wlr_output;
 	wlr_output->data = output;
 	output->server = server;
+	output->id_bit = id_bit;
 	output_state_init(output);
 
 	wl_list_insert(&server->outputs, &output->link);
@@ -500,7 +542,7 @@ handle_new_output(struct wl_listener *listener, void *data)
 	wl_signal_add(&wlr_output->events.request_state, &output->request_state);
 
 	wl_list_init(&output->regions);
-	wl_array_init(&output->osd_scene.items);
+	wl_list_init(&output->cycle_osd.items);
 
 	/*
 	 * Create layer-trees (background, bottom, top and overlay) and
@@ -511,7 +553,7 @@ handle_new_output(struct wl_listener *listener, void *data)
 			wlr_scene_tree_create(&server->scene->tree);
 	}
 	output->layer_popup_tree = wlr_scene_tree_create(&server->scene->tree);
-	output->osd_tree = wlr_scene_tree_create(&server->scene->tree);
+	output->cycle_osd_tree = wlr_scene_tree_create(&server->scene->tree);
 	output->session_lock_tree = wlr_scene_tree_create(&server->scene->tree);
 
 	/*
@@ -535,7 +577,7 @@ handle_new_output(struct wl_listener *listener, void *data)
 	wlr_scene_node_place_below(&output->layer_tree[3]->node, menu_node);
 	wlr_scene_node_place_below(&output->layer_popup_tree->node, menu_node);
 
-	wlr_scene_node_raise_to_top(&output->osd_tree->node);
+	wlr_scene_node_raise_to_top(&output->cycle_osd_tree->node);
 	wlr_scene_node_raise_to_top(&output->session_lock_tree->node);
 
 	/*
@@ -583,6 +625,7 @@ output_init(struct server *server)
 		server->output_layout);
 
 	wl_list_init(&server->outputs);
+	server->next_output_id_bit = (1 << 0);
 
 	output_manager_init(server);
 }
@@ -1144,3 +1187,15 @@ output_enable_adaptive_sync(struct output *output, bool enabled)
 			enabled ? "en" : "dis", output->wlr_output->name);
 	}
 }
+
+void
+output_set_has_fullscreen_view(struct output *output, bool has_fullscreen_view)
+{
+	if (rc.adaptive_sync != LAB_ADAPTIVE_SYNC_FULLSCREEN
+			|| !output_is_usable(output)) {
+		return;
+	}
+	/* Enable adaptive sync if view is fullscreen */
+	output_enable_adaptive_sync(output, has_fullscreen_view);
+	output_state_commit(output);
+}
